const path = require('path');
const pathSep = path.sep;
const fs = require('fs');

// 利用 handle 模板替换以下值
const PROJECT_ROOT = '{{ PROJECT_ROOT }}'
const FILTER_MODULES_PATH = '{{ FILTER_MODULES_PATH }}'
const IMAGE_BASE_URL = '{{ IMAGE_BASE_URL }}'
const MODE = '{{ MODE }}'
const USE_HY_CSS = {{ USE_HY_CSS }}
const CSS_TRANSFORMER_PATH = '{{ CSS_TRANSFORMER_PATH }}'

function defaultFactory() {
  const fileToIdMap = new Map();
  let nextId = 0;
  return modulePath => {
    let id = fileToIdMap.get(modulePath);
    if (typeof id !== 'number') {
      id = nextId++;
      fileToIdMap.set(modulePath, id);
    }
    return id;
  };
}

function getFilterModules(filePath) {
  if (fs.existsSync(filePath)) {
    return new Set(fs.readFileSync(filePath).toString().split('\n'));
  }

  return new Set()
}

function moduleIdFactory() {
  if (MODE === 'dev') {
    return defaultFactory()
  }

  return modulePath => {
    //RN JS 全部以 RN_ 开头
    if (modulePath.includes('node_modules' + pathSep + 'react-native' + pathSep + 'Libraries' + pathSep)) {
      return `RN_${path.basename(modulePath, '.js')}`
    }

    let name = modulePath.replace(`${PROJECT_ROOT}/`, '')

    name = name.replace('.js', '');

    //路径分隔符换成 _
    const regExp = pathSep == '\\' ? new RegExp('\\\\', 'g') : new RegExp(pathSep, 'g');
    name = name.replace(regExp, '_');

    //把业务/base jsbundle 里的非 node_modules 里的模块 id 都带上根目录名
    if (modulePath.includes(PROJECT_ROOT) && !modulePath.includes('node_modules')) {
      let basename = path.basename(PROJECT_ROOT)
      //小程序的 path.basename 需要替换成 open-sdk
      //线上运行的小程序全都是从 open-sdk require，所以这里只能继续目录名映射
      if (basename === 'hyext-rn-sdk' || basename === 'hyliveext-rn-sdk') {
        basename = 'open-sdk'
      }
      name = basename + '_' + name;
    }


    /*
     * SDK => open-sdk 只兼容了之前的虎牙小程序的名称, 没有兼容中台
     * 因为中台的都是62的rn版本
     */

    //用户端小程序打包
    const extRequireReg = /(node_modules_hyext-rn-sdk|node_modules_@hyext_hyext-rn-sdk)/ig
    if (extRequireReg.test(name)) {
      name = name.replace(extRequireReg, 'open-sdk')
    }


    //主播端小程序打包
    const liveExtRequireReg = /(node_modules_hyliveext-rn-sdk|node_modules_@hyext_hyliveext-rn-sdk)/ig
    if (liveExtRequireReg.test(name)) {
      name = name.replace(liveExtRequireReg, 'open-sdk')
      // console.log(`\n====主播端小程序打包, node_modules_hyliveext-rn-sdk => open-sdk : ${name}`)
    }

    return name;
  };
}



const filterModules = getFilterModules(FILTER_MODULES_PATH)
function processModuleFilter(module) {
  if (MODE === 'dev') {
    return true
  }

  const modulePath = module['path']

  // 这个文件baseBundle已经包含了, 如果多次打包, 运行的时候会报错
  if (modulePath.endsWith('/node_modules/metro/src/lib/polyfills/require.js')) {
    return false
  }

  const moduleId = moduleIdFactory()(modulePath)
  if (filterModules.has(moduleId)) {
    return false
  }


  if (modulePath.indexOf('__prelude__') >= 0) {
    if (process.env.HYEXT_RN_VERSION === '0.58.5' && process.env.HYEXT_RN_HOOK) {
      const hook = Buffer.from(process.env.HYEXT_RN_HOOK || '', 'base64')
      const hookText = hook.toString('ascii')
      console.log(`\nNEO: hook in env`, hookText)
      module.output[0]['data']['code'] = hookText
    }
    return true;
  }

  let code = module.output[0]['data']['code'];

  const nameMatch = code.match(/name[ ]*:[ ]*"(\S+?)"/);
  if (nameMatch && nameMatch.length >= 1) {
    const name = nameMatch[1];

    const typeMatch = code.match(/type[ ]*:[ ]*"(\S+?)"/);
    if (typeMatch && typeMatch.length >= 1) {
      const type = typeMatch[1];

      const relativeMatch = code.match(/httpServerLocation[ ]*:[ ]*"(\S+?)"/);
      if (relativeMatch && relativeMatch.length >= 1) {
        const relativePath = relativeMatch[1];

        //read image baseURI from process.env
        const uri = IMAGE_BASE_URL + relativePath + '/' + name + '.' + type
        code = code.replace(/\.exports=(\S+)}/, `.exports={uri:"${uri}"}}`)
        module.output[0]['data']['code'] = code
      }
    }
  }

  return true;
}

const basicExts = ['js', 'json', 'ts', 'tsx']

const basicConfig = {
  resolver: {
    platforms: [ 'ios', 'android', 'native', 'web' ],
    sourceExts: USE_HY_CSS ? [...basicExts, 'hycss', 'css'] : basicExts,
  },
  serializer: {
    createModuleIdFactory: moduleIdFactory,
    processModuleFilter: processModuleFilter
  },
  projectRoot: PROJECT_ROOT,
}

const hyCssConfig = {
  ...basicConfig,
  transformer: {
    babelTransformerPath: CSS_TRANSFORMER_PATH,
  },
}

const finnalConfig = USE_HY_CSS ? hyCssConfig : basicConfig

module.exports = finnalConfig



