import R from 'ramda'
import path from 'path'

export type RNExtType = 'app_panel' | 'app_popup' | 'zs_anchor_panel' | 'zs_anchor_popup' | 'mid_zs_panel' | 'mid_app_panel'

export type Platform = 'ios' | 'android'

export type BundleType = 'RN_BUNDLE' | 'MINI_GAME_RN' | string

export interface RNBuildResult {
  // 枚举值: RNExtType
  [extType: string]: {
    type: BundleType | string
    content: {
      engine?: 'h5game' | 'laya'

      // 如 '1.10.0'
      sdkVersion: string

      /*
       * 只有在开发环境才有用, 其他环境为 ''
       */
      baseURI: string

      rnBundle: {
        // 枚举值: Platform
        [platform: string]: {
          // 相对 publicPath 的路径, 如 rn/app_panel/app.ios.jsbundle
          path: string

          entries: Array<{
            // registerComponent name: 'App'
            name: string

            // rnext=1&hyaction=newrn&rnmodule=${}&rnentry=${}&sdkversion=${}&rnmd5=${rnmd5}
            url: string
          }>
        }
      }

      // 如果是小游戏类型, 必须存在
      gameSource?: Array<{
        // hyg file name, 如 rw20dt3d.hyg
        name: string
        // hyg link address, 如 https://hedd67cfa5c32ad1-rw2odt3d.hyext.com/des/0.3.0/1cd60dcacd32248a52cc885fbbba829c/f36e6aa15efc9084ee62910d565e97e3/rn/rw2odt3d.hyg
        path: string
        // bytes of hyg file
        size: number
        // md5 of hyg file
        md5: string
      }>
    }
  }
}


type GenerateUrlOption = {
  extType: string
  extUuid: string
  registerComponent: string
  SDKVersion: string
  md5: string
  rnbaseurl?: string
  rnbasemd5?: string
  RNVersion?: string
}
export function generateProdUrl({
  extType,
  extUuid,
  registerComponent,
  SDKVersion,
  md5,
  rnbaseurl,
  rnbasemd5,
  RNVersion,
}: GenerateUrlOption) {

  let values: { [key: string]: number | string } = {
    rnext: 1,
    hyaction: 'newrn',

    // 1. 不要轻易改变这个 rnmodule 格式, 可能会影响到后端的统计
    // 2. 这个名称不要包含非 ascii 值
    rnmodule: `${extUuid}-${extType}-${extUuid}`,
    rnentry: registerComponent,
    sdkversion: SDKVersion,
    rnmd5: md5,
    framework: RNVersion || process.env.HYEXT_RN_VER || '0',
  }

  if (rnbaseurl && rnbasemd5) {
    values = {
      ...values,
      rnbaseurl,
      rnbasemd5,
    }
  }

  return Object.entries(values).map(([k ,v]) => `${k}=${encodeURIComponent(v)}`).join('&')
}

const generateDevUrl = (component: string, platform: Platform) => {
  return `rnentry=${component}&platform=${platform}&rnext=1&hyaction=newrn`
}

interface BuildInfo {
  extType: RNExtType | string

  /*
   * 开发模式下: 类似 index.bundle
   * 其他模式下: 类似 rn/app_panel/app.android.jsbundle
   */
  bundleRelativePath: string

  /*
   * queryString: 下发给终端的url, 没有的话用默认的规则生成一个
   */
  registerComponents: { name: string, queryString?: string }[]
  platform: Platform
  SDKVersion: string
  bundleType?: BundleType
}
interface ProdBuildInfo extends BuildInfo {
  bundleMd5: string
  baseUrl?: string
  baseMd5?: string
}
interface UniformBuildInfo extends BuildInfo {
  registerComponents: { name: string, queryString: string }[]
}


export interface BaseOptions {
  builderName: string
}
interface ProdOptions extends BaseOptions {
  extInfo: {
    packageName: string
    extUuid: string
  }
  buildResultConfig: ProdBuildInfo[]
}
interface DevOptions extends BaseOptions {
  // 开发模式下, 为打包服务的地址, 如 https://172.168.1.1:8081
  devServerOrigin: string
  buildResultConfig: BuildInfo[]
}
interface UniformOptions extends BaseOptions {
  devServerOrigin: string
  buildResultConfig: UniformBuildInfo[]
}

export function generateProdBuildResult({ builderName, extInfo, buildResultConfig }: ProdOptions): RNBuildResult {
  const config = buildResultConfig.map((cfg: ProdBuildInfo) => ({
    ...cfg,
    registerComponents: cfg.registerComponents.map(({ name, queryString }) => ({
      name,
      queryString: queryString || generateProdUrl({
        ...extInfo,
        extType: cfg.extType,
        registerComponent: name,
        SDKVersion: cfg.SDKVersion,
        md5: cfg.bundleMd5,
        rnbaseurl: cfg.baseUrl,
        rnbasemd5: cfg.baseMd5,
      }),
    }))
  }))

  return generateBuildResult({
    builderName,
    devServerOrigin: '',
    buildResultConfig: config,
  })
}

export function generateDevBuildResult({ builderName, devServerOrigin, buildResultConfig }: DevOptions): RNBuildResult {
  const config = buildResultConfig.map((cfg: BuildInfo) => ({
    ...cfg,
    registerComponents: cfg.registerComponents.map(({ name, queryString }) => ({
      name,
      queryString: queryString || generateDevUrl(name, cfg.platform)
    }))
  }))

  return generateBuildResult({
    builderName,
    devServerOrigin,
    buildResultConfig: config,
  })
}


export function generateBuildResult({ builderName, devServerOrigin, buildResultConfig }: UniformOptions): RNBuildResult {
  let buildResult = {}

  buildResultConfig.forEach(({ extType, bundleRelativePath, registerComponents, platform, SDKVersion, bundleType = 'RN_BUNDLE' }) => {
    const result = {
      [extType]: {
        type: bundleType,
        builder: builderName,
        content: {
          sdkVersion: SDKVersion,
          baseURI: devServerOrigin,
          rnBundle: {
            [platform]: {
              path: bundleRelativePath,
              entries: registerComponents.map(({ name, queryString }) => ({
                name,
                url: queryString,
              }))
            },
          },
        },
      },
    }

    buildResult = R.mergeDeepLeft(
      result,
      buildResult,
    )
  })


  return buildResult
}
