import path from 'path'
import os from 'os'
import fs from 'fs-extra'
import crypto from 'crypto'
import Handlebars from 'handlebars'
// @ts-ignore untyped metro-cache
import { FileStore, AutoCleanFileStore } from 'metro-cache'

import { MetroConfig } from './configTypes'

// @ts-ignore
import { getDefaultConfig as getOfficialDefaultConfig, mergeConfig } from 'metro-config'
// @ts-ignore
import { getDefaultValues } from 'metro-config/src/defaults'
import { getDefaultConfig as getOfficialCliConfig } from '@react-native-community/cli-plugin-metro'

import { defaultFactory, createModuleIdHOF } from './createModuleIdHOF'
import resolveReactNativePath from './resolveReactNativePath'
import filterModuleHOF from './filterModuleHOF'

interface BaseOptions {
  projectRoot: string
  extUuid?: string
  cacheVersion?: string
  port?: number
}

interface DevOptions extends BaseOptions {
  mode: 'dev',
  imageBaseURL?: string
  SDKAlias?: { [aliasName: string]: string[] }
  filterModules?: Set<string>
  filterModulesVersion?: Set<string>
}

interface ProdOptions extends BaseOptions {
  mode: 'production'

  /*
   * 小程序打包时, 会把相对路径的图片引用改成对cdn图片的引用
   * 1. 如果打包时把图片保存到 outputDir, 那么 imageBaseURL 和 publicPath 相同
   * 2. 如果打包时把图片保存到 outputDir/app_panel/, 那么 imageBaseURL 和 publicPath/app_panel/ 相同
   *
   * 默认情况下, outputDir 是 publicPath. 除非在构建时 BundleConfig 显示指定了 assetsDir
   * 默认情况下, 图片资源会打包到跟 jsBundle 同级目录, 所以 imageBaseURL 一般是 publicPath/jsBundle_dir/
   *
   */
  imageBaseURL: string

  /*
   * SDK 的npm包名称, 包含 org 前缀(如有), 如 ['@hyext/hyext-rn-sdk'], ['hyext-rn-sdk']
   * 业务代码对 sdk 的引用将会被替换成 open-sdk(open-sdk 在app内预置了, 可直接使用)
   * 如 node_modules_@hyext_hyext-rn-sdk 将会变替换成 open-sdk
   */
  SDKAlias: { [aliasName: string]: string[] }

  filterModules: Set<string>
  filterModulesVersion: Set<string>
}


/*
 * SDKRole: 'streamer' | 'audience' 仅在线上打包使用:
 * 1. filterModuleHOF 用来过滤filterModules 里指定的包
 * 2. transform 用来替换sdk
 */
export default function getDefaultConfig(options: DevOptions | ProdOptions): MetroConfig {
  const {
    projectRoot,
    imageBaseURL = '',
    filterModules = new Set(),
    filterModulesVersion = new Set(),
    SDKAlias = {},
    mode = 'dev',
    extUuid = 'huya-ext-shared-uuid',
    cacheVersion = '1.0',
    port = 8081,
  } = options

  // metro 默认值: metro-config/src/defaults/index.js
  const officialDefaultConfig = getDefaultValues(projectRoot)

  // 
  const officialCliConfig = getOfficialCliConfig({
    platforms: {},
    reactNativePath: resolveReactNativePath(projectRoot),
  } as any)

  const officialConfig = mergeConfig(
    officialDefaultConfig,
    officialCliConfig,
  )

  const sepPos = __dirname.indexOf('/node_modules/')
  const topMostNodeModules = `${__dirname.slice(0, sepPos)}/node_modules`

  const CACHE_PATH = path.join(
    process.env.NODE_PACK_CACHE_ROOT || os.tmpdir(),
    'huya-ext-metro-cache',
    mode === 'dev'
      ? `${path.basename(projectRoot)}@${cacheVersion}`
      : `${extUuid}@${cacheVersion}`,
  )

  const hyextConfig = {
    resolver: {
      sourceExts: [...officialConfig.resolver.sourceExts, 'hycss'],
      platforms: ['ios', 'android', 'native', 'web'],
    },

    serializer: {
      processModuleFilter: filterModuleHOF({ projectRoot, imageBaseURL, SDKAlias, filterModules, filterModulesVersion, mode }),
      createModuleIdFactory: mode === 'dev' ? defaultFactory : createModuleIdHOF({ projectRoot, SDKAlias, }),
    },

    server: {
      port,

      // 中间件的执行顺序: communityCLI => metroMiddleware.
      // 这个中间件中的逻辑, 会晚于 communityCLI 的中间件执行
      enhanceMiddleware: undefined,
    },

    transformer: {
      babelPlugins: [],
      enableBabelRCLookup: false,
      getTransformOptions: async () => ({
        transform: {
          asyncRequireModulePath: require.resolve('metro-runtime/src/modules/asyncRequire'),
          babelTransformerPath: require.resolve('metro-react-native-babel-transformer'),
          experimentalImportSupport: false,

          /*
           * 开发环境下, 这个值设置为true, 否则会遇到问题:
           * https://github.com/facebook/react-native/issues/28801
           */
          inlineRequires: mode === 'dev',
          unstable_disableES6Transforms: false,
        },
        preloadedModules: false,
        ramGroups: [],
      }),

      // 这个地址不要轻易改动, 否则开发环境加载不出图片
      publicPath: mode === 'dev' ? '' : '/hyext-assets',
    },

    extUuid,

    cacheStores: [
      new FileStore({
        // 使用系统的临时文件夹, 由操作系统决定何时清除
        root: CACHE_PATH,
      })
    ],
    hasteMapCacheDirectory: CACHE_PATH,

    // 确保不同的小程序能产生不同的缓存值
    cacheVersion: mode === 'dev'
      ? `${path.basename(projectRoot)}@${cacheVersion}`
      : `${extUuid}@${cacheVersion}`,
    projectRoot,
    watchFolders: [projectRoot, topMostNodeModules], // 不监听 topMostNodeModules 的话, topMostNodeModules 里的一些包引入不到

    onReady: () => { },
    enableRootBabelRC: false,
  }

  return mergeConfig(
    officialDefaultConfig,
    officialCliConfig,
    hyextConfig,
  )
}

export function generateTransformerFile(options: {
  projectRoot: string
  extUuid?: string
  cacheVersion?: string
  babelPlugins?: any[]
  hyextTransformerConfig?: any
}) {
  const {
    projectRoot,
    extUuid = 'huya-ext-shared-uuid',
    cacheVersion = '1.0',
    babelPlugins = [],
    hyextTransformerConfig = {},
  } = options

  const identity = crypto
    .createHash('sha1')
    .update([extUuid || projectRoot, cacheVersion].join('$'))
    .digest('hex')

  const tempalteFile = path.join(__dirname, './babelTransformer.js.hbs')
  const destFile = path.join(__dirname, `${identity}.babelTransformer.js`)

  const template = Handlebars.compile(fs.readFileSync(tempalteFile, 'utf8'))
  fs.writeFileSync(
    destFile,
    template({
      HANDLE_BARS_PLUGINS: JSON.stringify(babelPlugins),
      HANDLE_BARS_HYEXT_TRANSFORMER_CONFIG: JSON.stringify(hyextTransformerConfig),
    }),
    {
      encoding: 'utf8',
    },
  )

  return destFile
}
